package model;

import Jama.Matrix;
import java.math.BigDecimal;
import java.math.RoundingMode;
import java.util.List;

/**
 * Třída pro reprezentaci vektoru s prvky ve formátu double
 *
 * @author Matěj Fries
 */
public class Vektor
{

    /**
     * Pole s prvky vektoru
     */
    private double[] prvky;

    /**
     * Konstruktor vektoru z pole prvků
     *
     * @param prvky pole s prvky vektoru ve formátu double
     */
    public Vektor(double[] prvky)
    {
        this.prvky = new double[prvky.length];
        for (int i = 0; i < prvky.length; i++)
        {
            this.prvky[i] = prvky[i];
        }
    }

    /**
     * Konstruktor vektoru z listu prkvů
     *
     * @param prvky list s prvky vektoru ve formátu double
     */
    public Vektor(List<Double> prvky)
    {
        this.prvky = new double[prvky.size()];
        for (int i = 0; i < prvky.size(); i++)
        {
            this.prvky[i] = prvky.get(i);
        }
    }

    /**
     * Vrací prvky vektoru
     *
     * @return pole ve formátu double s prvky vektoru
     */
    public double[] getPrvky()
    {
        return prvky;
    }

    /**
     * Vypočítá velikost vektoru
     *
     * @return vrací číslo ve formátu double
     */
    public double velikostVektoru()
    {
        double velikost = 0.0;
        for (int i = 0; i < prvky.length; i++)
        {
            velikost += prvky[i] * prvky[i];
        }
        velikost = Math.sqrt(velikost);
        return velikost;
    }

    /**
     * Vypočítá součet prvků vektoru
     *
     * @return součet prvků vektoru
     */
    public double soucetVektoru()
    {
        double soucet = 0.0;
        for (int i = 0; i < prvky.length; i++)
        {
            soucet += prvky[i];
        }

        return soucet;
    }

    /**
     * Vrátí nový normalizovaný vektor
     *
     * @return vrací normalizovaný vektor
     */
    public Vektor normalizovanyVektor()
    {
        double velikost = this.velikostVektoru();
        if (velikost > 0)
        {
            double[] normalizovanePrvky = new double[this.getPrvky().length];

            for (int i = 0; i < this.getPrvky().length; i++)
            {
                normalizovanePrvky[i] = this.getPrvek(i) / velikost;
            }

            return new Vektor(normalizovanePrvky);
        } else
        {
            return new Vektor(this.getPrvky());
        }

    }

    /**
     * Pomocná metoda pro konzolový výstup
     */
    public void print()
    {
        System.out.println();

        for (int i = 0; i < this.getPrvky().length; i++)
        {
            System.out.print(this.getPrvek(i) + " ");
        }

        System.out.println();
        System.out.println();
    }

    /**
     * Přičte k vektoru druhý vektor
     *
     * @param vektor vektor k přičtení
     */
    public void plusVektor(Vektor vektor)
    {
        if (this.getPrvky().length != vektor.getPrvky().length)
        {
            throw new IllegalArgumentException("Vektory musí mít stejný počet prvků.");
        }

        for (int i = 0; i < prvky.length; i++)
        {
            prvky[i] = prvky[i] + vektor.getPrvek(i);
        }
    }

    /**
     * Vynásobí vektor číslem
     *
     * @param cislo číslo k vynásobení
     */
    public void kratCislo(double cislo)
    {
        for (int i = 0; i < prvky.length; i++)
        {
            prvky[i] = prvky[i] * cislo;
        }
    }

    /**
     * Odečte od vektoru druhý vektor
     *
     * @param vektor vektor k odečtení
     */
    public void minusVektor(Vektor vektor)
    {
        if (prvky.length != vektor.getPrvky().length)
        {
            throw new IllegalArgumentException("Vektory musí mít stejný počet prvků");
        }
        vektor.kratCislo(-1);
        this.plusVektor(vektor);
    }

    /**
     * Vynásobí vektor druhým vektorem
     *
     * @param vektor vektor k násobení
     */
    public void kratVektor(Vektor vektor)
    {
        if (prvky.length != vektor.getPrvky().length)
        {
            throw new IllegalArgumentException("Vektory musí mít stejný počet prvků");
        }

        for (int i = 0; i < prvky.length; i++)
        {
            prvky[i] = prvky[i] * vektor.getPrvek(i);
        }
    }

    /**
     * Metoda pro vypočítání skalárního součinu vektorů
     *
     * @param a první vektor
     * @param b druhý vektor
     * @return skalární součin vektorů ve formátu double
     */
    public static double skalarniSoucin(Vektor a, Vektor b)
    {
        if (a.getPrvky().length != b.getPrvky().length)
        {
            throw new IllegalArgumentException("Vektory musí mít stejný počet prvků");
        }
        double soucin = 0.0;

        for (int i = 0; i < a.getPrvky().length; i++)
        {
            soucin += a.getPrvky()[i] * b.getPrvky()[i];
        }

        return soucin;
    }

    /**
     * Metoda pro násobení matice sloupcovým vektorem
     *
     * @param v sloupcový vektor
     * @param mat matice
     * @return nový vektor vznikly součinem vektoru a matice
     */
    public static Vektor nasobeniMaticeVektorem(Vektor v, Matrix mat)
    {
        if (mat.getColumnDimension() != v.getPrvky().length)
        {
            throw new IllegalArgumentException("Vektoru musí mít stejně prvků jako matice sloupců.");
        }
        double[] p = new double[mat.getRowDimension()];
        for (int i = 0; i < p.length; i++)
        {
            double soucet = 0.0;
            for (int j = 0; j < mat.getColumnDimension(); j++)
            {
                soucet += mat.get(i, j) * v.getPrvky()[j];
            }
            p[i] = soucet;
        }

        Vektor nas = new Vektor(p);
        return nas;
    }

    /**
     * Metoda pro vypočítání cosinovy podobnosti vektorů
     *
     * @param a prnví vektor
     * @param b druhý vektor
     * @return číslo ve formátu double udávající podobnost vektorů (od -1 do 1)
     */
    public static double cosinovaPodobnostVektoru(Vektor a, Vektor b)
    {
        if (a.getPrvky().length != b.getPrvky().length)
        {
            throw new IllegalArgumentException("Vektory musí mít stejný počet prvků");
        }
        double soucin;
        double soucinVelikosti;
        double podobnost;

        soucin = Vektor.skalarniSoucin(a, b);
        soucinVelikosti = a.velikostVektoru() * b.velikostVektoru();
        podobnost = soucin / soucinVelikosti;

        return podobnost;
    }

    /**
     * Nastavuje vektoru nové hodnoty prvků
     *
     * @param prvky pole nových hodnot prvků ve formátu double
     */
    public void setPrvky(double[] prvky)
    {
        this.prvky = prvky;
    }

    /**
     * Nastavuje novou hodnotu jednoho prvku
     *
     * @param hodnota nová hodnota prvky ve formátu double
     * @param pozice pozice prvku ve vektoru
     */
    public void setPrvek(double hodnota, int pozice)
    {
        prvky[pozice] = hodnota;
    }

    /**
     * Vrací hodnotu prvku vektoru
     *
     * @param pozice pozice prvku ve vektoru
     * @return hodnota prvku ve formátu double
     */
    public double getPrvek(int pozice)
    {
        return prvky[pozice];
    }

    /**
     * Metoda pro prohození řádků nebo sloupců matice
     *
     * @param a číslo prvního řádku nebo sloupce k prohození
     * @param b číslo druhého řádku nebo sloupce k prohození
     * @param radky přepínač, pokud je true, prohazují se řádky, pokud false,
     * prohazují se sloupce
     * @param matice matice
     * @return vrací novou matici s přehozenými vektory
     */
    public static Matice prohodVektory(int a, int b, boolean radky, Matice matice)
    {
        if (a < 0 || b < 0)
        {
            throw new IllegalArgumentException("Čísla řádků nebo sloupců musí být větší než nula.");
        }
        Matice mat = new Matice(matice.getArrayCopy());
        if (a != b)
        {
            double m = mat.getRowDimension();
            double n = mat.getColumnDimension();
            if (radky)
            {
                if (a > m || b > m)
                {
                    throw new IllegalArgumentException("Čísla řádků jsou větší než počet řádků v matici.");
                }
                for (int i = 0; i < n; i++)
                {
                    double pom = mat.get(a, i);
                    mat.set(a, i, mat.get(b, i));
                    mat.set(b, i, pom);
                }
            } else
            {
                if (a > n || b > n)
                {
                    throw new IllegalArgumentException("Čísla sloupců jsou větší než počet sloupců v matici.");
                }
                for (int i = 0; i < m; i++)
                {
                    double pom = mat.get(i, a);
                    mat.set(i, a, mat.get(i, b));
                    mat.set(i, b, pom);
                }
            }
        }

        return mat;
    }

    /**
     * Metoda pro zaokrouhlení hodnot ve vektoru
     *
     * @param pocetMist počet desetinných míst k zaokrouhlení
     * @return vrací zaokrouhlený vektor
     */
    public Vektor zaokrouhliVektor(int pocetMist)
    {
        double[] hodnoty = this.getPrvky().clone();

        for (int x = 0; x < hodnoty.length; x++)
        {
            hodnoty[x] = Vektor.zaokrouhli(hodnoty[x], pocetMist);
        }

        return new Vektor(hodnoty);
    }

    /**
     * Metoda pro zaokrouhlení čísla
     *
     * @param hodnota číslo k zaokrouhlení
     * @param mist počet desetinných míst k zaokrouhlení
     * @return vrací zaokrouhlené číslo
     */
    public static double zaokrouhli(double hodnota, int mist)
    {
        if (mist > 0)
        {
            BigDecimal bd = new BigDecimal(hodnota);
            bd = bd.setScale(mist, RoundingMode.HALF_UP);
            return bd.doubleValue();
        }

        return hodnota;
    }
}
